import json

from moontour_common.models.rooms.game_modes.health_room import HealthRoom
from moontour_common.models.rooms.game_modes.race_room import RaceRoom
from moontour_common.models.rooms.game_modes.streak_room import StreakRoom
from moontour_common.models.rooms.participants.player import Player
from moontour_common.models.rooms.participants.team import Team
from moontour_common.models.rooms.room import Room

ROOM_MODELS: dict[str, type[Room]] = {
    cls.__name__: cls for cls in
    {HealthRoom[Player], HealthRoom[Team], RaceRoom[Player], RaceRoom[Team], StreakRoom}
}


def get_room_model(type_: str) -> type[Room]:
    return ROOM_MODELS[type_]


def parse_room(room_raw: str) -> Room:
    room_dict = json.loads(room_raw)
    room_model = get_room_model(room_dict['type'])
    return room_model.parse_obj(room_dict)


async def get_room(id_: str) -> Room:
    room_raw = await Room.commands.get(id_)
    return parse_room(room_raw)



from pydantic import BaseModel

from moontour_common.models.coordinates import Coordinates
from moontour_common.models.guess import Guess
from moontour_common.models.rooms.participants.player import Player
from moontour_common.models.rooms.phase import Phase
from moontour_common.models.rooms.room import Room


class StreaksPhase(Phase):
    guess: Guess | None = None

    def insert_guess(self, user_id: str, coordinates: Coordinates):
        self.guess = Guess(coordinates=coordinates)


class ParticipantState(BaseModel):
    streak: int = 0


class StreakRoom(Room[StreaksPhase, Player, ParticipantState]):
    pass



from datetime import datetime
from enum import StrEnum

from pydantic import BaseModel, Field

from moontour_common.models.coordinates import Coordinates
from moontour_common.models.guess import Guess
from moontour_common.models.location import Location


class PhaseStatus(StrEnum):
    waiting = 'waiting'
    running = 'running'
    finished = 'finished'


class Phase(BaseModel):
    target: Location
    status: PhaseStatus = PhaseStatus.waiting
    create_time: datetime = Field(default_factory=datetime.now)
    guesses: dict[str, Guess] = {}  # User ID to guess

    def insert_guess(self, user_id: str, guess: Guess):
        self.guesses[user_id] = guess



from typing import Generic

from pydantic import BaseModel

from moontour_common.models.rooms.phase import Phase
from moontour_common.models.rooms.room import Room
from moontour_common.models.rooms.room_model import ParticipantType

START_HEALTH = 5000


class HealthPhase(Phase):
    damage_multiplier: float = 1


class ParticipantState(BaseModel):
    health: int = START_HEALTH


class HealthRoom(Room[HealthPhase, ParticipantType, ParticipantState], Generic[ParticipantType]):
    guess_duration: float = 15  # Time between first guess to phase ending

    def should_close(self) -> bool:
        dead_count = len([participant for participant in self.participants if participant.state.health == 0])
        return dead_count == len(self.participants) - 1

    def _finalize_phase(self, phase: HealthPhase):
        guesses = phase.guesses
        best_guess = max([guess.compute_points(phase.target) for guess in guesses.values()]) if guesses else 0
        for participant in self.participants:
            participant_guesses = {
                guesses[user_id].compute_points(phase.target)
                for user_id in participant.get_user_ids() if user_id in guesses
            }
            best_participant_guess = max(participant_guesses) if participant_guesses else 0
            damage = best_guess - best_participant_guess
            participant.state.health = max(participant.state.health - damage, 0)



from abc import abstractmethod, ABC
from collections.abc import Awaitable
from contextlib import asynccontextmanager
from typing import TypeVar, Callable, ParamSpec, Generic

from aio_pika import Message

from moontour_common.commands.redis_room_commands import RedisRoomCommands
from moontour_common.database.rabbitmq import get_rooms_exchange
from moontour_common.models.coordinates import Coordinates
from moontour_common.models.guess import Guess
from moontour_common.models.rooms.phase import PhaseStatus
from moontour_common.models.rooms.room_model import RoomModel, RoomStatus, PhaseType, ParticipantType
from moontour_common.models.user import User

RoomType = TypeVar('RoomType', bound='Room')

P = ParamSpec('P')
T = TypeVar('T')


def modify_room(func: Callable[P, T]) -> Callable[P, Awaitable[T]]:
    async def wrapper(*args, **kwargs):
        room: 'Room' = args[0]
        async with room.modify():
            return func(*args, **kwargs)

    return wrapper


class Room(RoomModel[PhaseType, ParticipantType], ABC, Generic[PhaseType, ParticipantType]):
    commands = RedisRoomCommands()

    @classmethod
    async def get(cls: type[RoomType], id_: str) -> RoomType:
        room_raw = await Room.commands.get(id_)
        return cls.parse_raw(room_raw)

    @classmethod
    async def find(cls: type[RoomType]) -> RoomType | None:
        room_raw = await Room.commands.find(cls.__name__)
        if room_raw is not None:
            return cls.parse_raw(room_raw)

    @staticmethod
    async def delete(id_: str):
        await Room.commands.delete(id_)

    async def reload(self):
        new_room = self.get(self.id)
        self.__dict__.update(new_room.__dict__)

    async def save(self):
        await self.commands.save(self.id, self.json())

    async def notify(self):
        exchange = await get_rooms_exchange()
        await exchange.publish(
            message=Message(body=self.json().encode()),
            routing_key=f'{self.id}.*',
        )

    @asynccontextmanager
    async def modify(self):
        async with self.commands.lock():
            await self.reload()
            try:
                yield
            except Exception:
                raise
            else:
                await self.save()
                await self.notify()

    def _get_participant(self, user_id: str) -> ParticipantType | None:
        for participant in self.participants:
            if user_id in participant.get_user_ids():
                return participant

    @modify_room
    def join(self, user: User):
        if self.status != RoomStatus.waiting:
            return
        for participant in self.participants:
            if not participant.is_full():
                participant.insert_user(user)
                break
        if all([participant.is_full() for participant in self.participants]):
            self.status = RoomStatus.waiting_to_start

    @modify_room
    def leave(self, user_id: str):
        for participant in self.participants:
            participant.remove_user(user_id)

    @modify_room
    def start(self):
        self.status = RoomStatus.running

    @modify_room
    def add_phase(self, phase: PhaseType):
        self.phases.append(phase)

    @modify_room
    def start_phase(self):
        self.phases[-1].status = PhaseStatus.running

    @modify_room
    def guess(self, user_id: str, coordinates: Coordinates):
        phase = self.phases[-1]
        if self.status == RoomStatus.running and phase.status == PhaseStatus.running:
            guess = Guess(coordinates=coordinates)
            phase.insert_guess(user_id, guess)
            self._on_guess(user_id, guess)

    def _on_guess(self, user_id: str, guess: Guess):
        pass

    @modify_room
    def finish_phase(self):
        phase = self.phases[-1]
        phase.status = PhaseStatus.finished
        self._finalize_phase(phase)

    def _finalize_phase(self, phase: PhaseType):
        pass

    @abstractmethod
    def should_close(self) -> bool:
        raise NotImplementedError()

    @modify_room
    def close(self):
        self.status = RoomStatus.closed



from moontour_common.models.coordinates import Coordinates
from moontour_common.models.game_messages.game_message import GameMessage


class GuessMessage(GameMessage):
    coordinates: Coordinates
    user_id: str



from datetime import datetime

import geopy.distance
from pydantic import BaseModel, Field

from moontour_common.models.coordinates import Coordinates

MAX_POINTS = 5000


class Guess(BaseModel):
    time: datetime = Field(default_factory=datetime.now)
    coordinates: Coordinates

    def compute_distance(self, target: Coordinates) -> float:
        return geopy.distance.geodesic(
            (target.latitude, target.longitude),
            (self.coordinates.latitude, self.coordinates.longitude)
        ).km

    def compute_points(self, target: Coordinates) -> int:
        distance = self.compute_distance(target)
        return max(MAX_POINTS - int(distance), 0)



from abc import abstractmethod
from contextlib import asynccontextmanager
from typing import Protocol


class BaseRoomCommands(Protocol):
    @abstractmethod
    async def get(self, id_: str) -> str: ...

    @abstractmethod
    async def find(self, type_: str) -> str | None: ...

    @abstractmethod
    async def delete(self, id_: str): ...

    @abstractmethod
    async def save(self, id_: str, json_data: str): ...

    @asynccontextmanager
    async def lock(self, id_: str): ...



from abc import ABC, abstractmethod
from random import randint
from typing import TypeVar, Generic, Protocol

from pydantic import Field
from pydantic.color import Color
from pydantic.generics import GenericModel

from moontour_common.models.user import User


def get_random_color() -> Color:
    def get_component() -> int:
        return randint(0, 255)

    return Color((get_component(), get_component(), get_component()))


class Initializable(Protocol):
    def __init__(self): ...


ParticipantState = TypeVar('ParticipantState', bound=Initializable)


class Participant(GenericModel, Generic[ParticipantState], ABC):
    color: Color = Field(default_factory=get_random_color)
    state: ParticipantState

    def __init__(self, **kwargs):
        state_type: type[ParticipantState] = self.__fields__['state'].type_
        super().__init__(state=state_type(), **kwargs)

    @abstractmethod
    def insert_user(self, user: User):
        raise NotImplementedError()

    @abstractmethod
    def remove_user(self, user_id: str):
        raise NotImplementedError()

    @abstractmethod
    def is_full(self) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def get_user_ids(self) -> set[str]:
        raise NotImplementedError()



import uuid
from datetime import datetime
from enum import StrEnum
from typing import Generic, TypeVar

from pydantic import Field, validator
from pydantic.generics import GenericModel

from moontour_common.models.rooms.participants.participant import Participant
from moontour_common.models.rooms.phase import Phase


class RoomStatus(StrEnum):
    waiting = 'waiting'
    waiting_to_start = 'waitingToStart'
    running = 'running'
    closed = 'closed'


PhaseType = TypeVar('PhaseType', bound=Phase)
ParticipantType = TypeVar('ParticipantType', bound=Participant)


class RoomModel(GenericModel, Generic[PhaseType, ParticipantType]):
    id: str = Field(alias='id_', default_factory=lambda: str(uuid.uuid4()))
    type: str = None
    map: str = 'world'
    status: RoomStatus = RoomStatus.waiting
    create_time: datetime = Field(default_factory=datetime.now)
    phases: list[PhaseType] = []
    participants: list[ParticipantType]

    def is_full(self):
        return all([participant.is_full() for participant in self.participants])

    @validator("type", always=True)
    def compute_type(cls):
        return cls.__name__

    class Config:
        allow_population_by_field_name = True



from contextlib import asynccontextmanager
from datetime import timedelta

from redis.commands.json.path import Path
from redis.commands.search.field import TextField
from redis.commands.search.indexDefinition import IndexDefinition, IndexType
from redis.exceptions import ResponseError

from moontour_common.commands.base_room_commands import BaseRoomCommands
from moontour_common.database.redis import redis_client
from moontour_common.models.rooms.room import RoomStatus

KEY_PREFIX = 'room:'
ROOMS_INDEX = 'rooms_idx'


def get_room_key(id_: str) -> str:
    return f'{KEY_PREFIX}{id_}'


class RedisRoomCommands(BaseRoomCommands):
    async def get(self, id_: str) -> str:
        return await redis_client.execute_command('JSON.GET', get_room_key(id_), Path.root_path())

    async def find(self, type_: str) -> str | None:
        result = await redis_client.ft(ROOMS_INDEX).search(
            f'@type:{type_} @status:{RoomStatus.waiting}')

        if len(result.docs) > 0:
            return result.docs[0].json

    async def delete(self, id_: str):
        await redis_client.execute_command('JSON.DEL', get_room_key(id_), Path.root_path())

    async def save(self, id_: str, json_data: str):
        await redis_client.execute_command('JSON.SET', get_room_key(id_), Path.root_path(), json_data)

    @asynccontextmanager
    async def lock(self, id_: str):
        async with redis_client.lock(f'room-{id_}'):
            yield



from typing import Generic

from pydantic import BaseModel

from moontour_common.models.guess import Guess
from moontour_common.models.rooms.phase import Phase
from moontour_common.models.rooms.room import Room, modify_room
from moontour_common.models.rooms.room_model import ParticipantType


class ParticipantState(BaseModel):
    current_phase: int = 0
    stunned: bool = False


class RaceRoom(Room[Phase, ParticipantType, ParticipantState], Generic[ParticipantType]):
    phase_count: int = 10
    stun_duration: float = 5

    @modify_room
    def initialize_phases(self, phases: list[Phase]):
        self.phases = phases

    def _on_guess(self, user_id: str, guess: Guess):
        participant = self._get_participant(user_id)
        target = self.phases[participant.state.current_phase].target
        if guess.compute_points(target) > 4000:
            participant.state.current_phase += 1
        else:
            participant.state.stunned = True

    def should_close(self) -> bool:
        return any([participant.state.current_phase == self.phase_count for participant in self.participants])



from pydantic import BaseModel, validator


class GameMessage(BaseModel):
    room_id: str
    type: str = None

    @validator("type", always=True)
    def compute_type(cls):
        return cls.__name__



from typing import Generic

from moontour_common.models.rooms.participants.participant import Participant, ParticipantState
from moontour_common.models.user import User


class Player(Participant[ParticipantState], Generic[ParticipantState]):
    user: User | None = None

    def insert_user(self, user: User):
        self.user = user

    def remove_user(self, user_id: str):
        if self.user.id == user_id:
            self.user = None

    def is_full(self) -> bool:
        return self.user is not None

    def get_user_ids(self) -> set[str]:
        return {self.user.id} if self.user else set()



import os

import aio_pika
from aio_pika import Message
from aio_pika.abc import AbstractChannel, AbstractExchange, AbstractQueue, ExchangeType

from moontour_common.models.game_messages.game_message import GameMessage

RABBITMQ_HOST = os.getenv('RABBITMQ_HOST', 'rabbitmq')
ROOMS_EXCHANGE_NAME = 'rooms'
GAME_EXCHANGE_NAME = 'game'
GAME_QUEUE_NAME = 'game'

_channel: AbstractChannel | None = None


async def get_rabbitmq_client() -> AbstractChannel:
    global _channel
    if _channel is None:
        connection = await aio_pika.connect_robust(host=RABBITMQ_HOST)
        _channel = await connection.channel()
    return _channel


async def get_rooms_exchange() -> AbstractExchange:
    client = await get_rabbitmq_client()
    return await client.declare_exchange(name=ROOMS_EXCHANGE_NAME, type=ExchangeType.TOPIC)


async def declare_player_queue(room_id: str, user_id: str) -> AbstractQueue:
    client = await get_rabbitmq_client()
    queue_name = f'{room_id}.{user_id}'
    queue = await client.declare_queue(name=queue_name, auto_delete=True)
    exchange = await get_rooms_exchange()
    await queue.bind(exchange, routing_key=f'{room_id}.*')
    return queue


async def get_game_exchange() -> AbstractExchange:
    client = await get_rabbitmq_client()
    return await client.declare_exchange(name='x-delayed-message', type=ExchangeType.X_DELAYED_MESSAGE)


async def declare_game_queue() -> AbstractQueue:
    client = await get_rabbitmq_client()
    queue = await client.declare_queue(name=GAME_QUEUE_NAME)
    exchange = await get_game_exchange()
    await queue.bind(exchange, routing_key=GAME_QUEUE_NAME)
    return queue


async def send_game_message(game_message: GameMessage, delay: float = 0):
    exchange = await get_game_exchange()
    await exchange.publish(
        Message(body=game_message.json().encode(), headers={'x-delay': 1000 * delay}),
        routing_key=GAME_QUEUE_NAME
    )



from typing import Generic

from moontour_common.models.rooms.participants.participant import Participant, ParticipantState
from moontour_common.models.user import User


class Team(Participant[ParticipantState], Generic[ParticipantState]):
    users: list[User] = []
    size: int = 2

    def insert_user(self, user: User):
        self.users.append(user)

    def remove_user(self, user_id: str):
        self.users = [user for user in self.users if user.id != user_id]

    def is_full(self) -> bool:
        return len(self.users) >= self.size

    def get_user_ids(self) -> set[str]:
        return {user.id for user in self.users}
